#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <strstream>
using namespace std;

struct OptionsType
{
	int out_start;
    int bytes_per_line;
} options;


void addchecksum(vector<unsigned char> &outbuffer);
char* hex(unsigned char c);
void outhex(ostream& out,vector<unsigned char> &outbuffer);
int bin2line(vector<unsigned char> &inbuffer, int first, int outaddress, vector<unsigned char> &outbuffer, int length=options.bytes_per_line);
int bin2hex(istream &in, ostream &out);
void outrecord(ostream& out, unsigned char recordtype, unsigned char size, unsigned char *data);

//int bin2hex(ostream &out,int page)
//{
//	
//	
//}

void addchecksum(vector<unsigned char> &outbuffer)
{
	unsigned char checksum=0;
	int first=0,last=outbuffer.size();
	int pos;
	for (pos=first;pos<last;pos++)
	{
		checksum+=outbuffer[pos];
	}
	checksum=-checksum;
	outbuffer.push_back(checksum);
}

char* hex(unsigned char c)
{
	static char hextable[17]="0123456789ABCDEF";
	static char hexval[3];
	hexval[0]=hextable[c>>4];
	hexval[1]=hextable[c&0xF];
	hexval[2]='\0';
	return hexval;
}

void outhex(ostream& out,vector<unsigned char> &outbuffer)
{
	int first=0, last;
	int pos;
	//add checksum
	addchecksum(outbuffer);
	
	last=outbuffer.size();
	//display colon
	out << ":";
	//display hex line
	for (pos=first;pos<last;pos++)
	{
		out << hex(outbuffer[pos]);
	}
	//end of line
	out << endl;
}

void addlineheader(vector <unsigned char> &outbuffer,unsigned char length, unsigned short address, unsigned char recordtype)
{
	outbuffer.push_back(length);
	outbuffer.push_back(unsigned char(address>>8));
	outbuffer.push_back(unsigned char(address&0xFF));
	outbuffer.push_back(recordtype);
}

void outrecord(ostream& out, unsigned char recordtype, unsigned char size, unsigned char *data)
{
	vector<unsigned char> outbuffer;
	addlineheader(outbuffer,size,0,recordtype);
	for (int pos=0;pos<size;pos++)
	{
		outbuffer.push_back(data[pos]);
	}
	outhex(out,outbuffer);
}


int bin2line(vector<unsigned char> &inbuffer, int first, int outaddress, vector<unsigned char> &outbuffer, int length)
//returns next address
{
	int inaddress;
	int last=first+length;
	if (last > inbuffer.size())
	{
		last=inbuffer.size();
		length=last-first;
	}
	
	addlineheader(outbuffer,length,outaddress,0);
	
	for (inaddress=first; inaddress<last;inaddress++)
	{
		outbuffer.push_back(inbuffer[inaddress]);
	}
	return last;
}

int bin2hex(istream &in, ostream &out)
{
	vector<unsigned char> inbuffer;
	vector<unsigned char> outbuffer;
	int insize;
	int address;
	
	//get size
	in.seekg(0,ios::end);
	insize=in.tellg();
	in.seekg(0,ios::beg);
	
	//allocate and read file
	inbuffer.resize(insize);
	in.read((char*)(&inbuffer[0]),insize);
	
	address=0;
	while (address<insize)
	{
		outbuffer.clear();
		address=bin2line(inbuffer,address,address+options.out_start,outbuffer);
		outhex(out,outbuffer);
	}

	return 0;
}


int getnum(char* arg)
{
	int retval=0;
	strstream in(arg,strlen(arg));
	in >> retval;
	return retval;
}

int main( int argc, char *argv[ ], char *envp[ ] )
{
	int arg;
	ostream* out=&cout;
	
	
	options.out_start=16384;
	options.bytes_per_line=16;

	ifstream fin;
	ofstream fout;
	if (argc>1)
	{
		for (arg=1;arg<argc;arg++)
		{
			if (argv[arg][0]=='-')
			{
				switch (char(tolower(argv[arg][1])))
				{
				case 'p':
					{
						unsigned short page=getnum(argv[arg]+2);
						unsigned char pagenum[2];
						pagenum[0]=page>>8;
						pagenum[1]=page&0xFF;
						outrecord(*out,2,2,pagenum);
					}
					break;
				case 'l':
					{
						options.bytes_per_line=getnum(argv[arg]+2);
					}
					break;
				case 's':
					{
						options.out_start=getnum(argv[arg]+2);
					}
					break;
				case 'e':
					{
						outrecord(*out,3,0,NULL);
						outrecord(*out,1,0,NULL);
					}
					break;
				case 'o':
					{
						++arg;
						if (arg>=argc) return 1;
						if (out==&fout)
						{
							fout.flush();
							fout.close();
							fout.clear();
						}
						fout.open(argv[arg]);
						out=&fout;
					}
					break;
				case 'c':
					{
						if (out==&fout)
						{
							fout.flush();
							fout.close();
							fout.clear();
						}
						out=&cout;
					}
					break;
				default:
					{
						cerr << "Unknown option: -"<<argv[arg][1] << endl;
					}
					break;
				}
			}
			else
			{
				fin.open(argv[arg],ios::binary);
				if (fin.good())
				{
					bin2hex(fin,*out);
				}
				fin.close();
				fin.clear();
			}
		}
	} else {
		cerr << "Syntax:" << endl;
		cerr << argv[0] << " [option or input filename]..." << endl;
		cerr << "options:" << endl;
		cerr << " -pN: set page number N" << endl;
		cerr << " -lN: set characters per line (default 16)" << endl;
		cerr << " -sN: set start address (default 0x4000)" << endl;
		cerr << " -e: write end record" << endl;
		cerr << " -o filename: output to filename" << endl;
		cerr << " -c: output to cout (default)" << endl;
		cerr << " filename: binary input from filename" << endl;
		cerr << endl;
		cerr << "Example: "<<argv[0]<<" -o bubapp.hex -p0 bubapp.bin -p1 bubapp2.bin -e" << endl;
		
		
	}

	if (out==&fout)
	{
		fout.flush();
		fout.close();
		fout.clear();
	}
	return 0;
}

